##
## jabberif jabber input plugin
##	this plugin provides jabber input support.
##	originally the jabber communication was the core part of this program.
##
##	written by Ben Fuhrmannek <bef@erlangen.ccc.de>
##

from baseclasses import input_plugin
from helpers import message
from string import split,strip,join
import re
import jabber
import socket

class jabber_input_plugin(input_plugin):
	def __init__(self, ph):
		self.ph = ph
		self._name = 'jabber_input_plugin'
		self._desc = 'jabber access'
		self._inputtype = 'jabber'

		## jabber users db as user/statustype key/value pairs
		self.__users = {}

		self.__default_config = {	'server':'jabber.org',
						'ssl':'on',
						'port':'5223',
						'username':'',
						'password':'',
						'resource':'jabberif',
						'priority':'3' }
		self._configure()

	def _configure(self):
		self.__config = self.ph.get_plugin_config(self, self.__default_config)

	def _on_load(self):
		self.connect()
		self.ph._register_command(self, 'julist', self.show_users, 'list jabber users', {'level':10})
		self.ph._register_command(self, 'jureload', self.reload_users, 'reload roaster', {'level':10})
		self.ph._register_command(self, 'jsendpr', self.send_presence, '[jid, type] send presence', {'level':10})
		self.ph._register_command(self, 'jsubscribe', self.send_subscribe, '[jid] send subscribe', {'level':10})

	def _on_unload(self):
		if hasattr(self, '__con'):
			self.__con.disconnect()
		
	####################################################################################################
	## callback functions
	
	def messageCB(self, con, msg):
		"""Called when a message is recieved"""
		if msg.getBody():
			msgbody = strip(msg.getBody())
			if len(msgbody) > 0:
				msg = message(self._inputtype, self, msg.getFrom().getStripped(), msgbody, args={'msg':msg})
				self.ph._process_request(msg=msg, threaded=True, dos_detection=True)

	def presenceCB(self, con, prs):
		"""Called when a presence is recieved"""
		who = prs.getFrom()
		type = prs.getType()
		if type == None: type = 'available'

		elif type == 'unsubscribe':
			con.send(jabber.Presence(to=str(who), type='unsubscribed'))
			con.send(jabber.Presence(to=str(who), type='unsubscribe'))

		elif type == 'subscribe':
			con.send(jabber.Presence(to=str(who), type='subscribe'))
		
		## update jabber users db
		self.__users[str(who)] = type


	def iqCB(self, con, iq):
		"""Called when an iq is recieved, we just let the library handle it at the moment"""
		pass

	def disconnectedCB(self, con):
		lh.log("Ouch, network error", 'error')
		## TODO: .. reconnect code missing here!
		raise Exception('unexpected disconnect')
		
	####################################################################################################
	## connection
	
	def connect(self):
		connection = 1
		if self.__config['ssl'] in ['1', 'on', 'yes']:
			connection = jabber.xmlstream.TCP_SSL
		self.__con = jabber.Client(host=self.__config['server'], port=int(self.__config['port']), connection=connection)

		try:
			self.__con.connect()
		except IOError, e:
			error = self.ph.log("Couldn't connect: %s" % e, 'error')
			return error
		else:
			self.ph.log("Connected",'connection')

		## handler registration
		self.__con.registerHandler('message', self.messageCB)
		self.__con.registerHandler('presence', self.presenceCB)
		self.__con.registerHandler('iq', self.iqCB)
		self.__con.setDisconnectHandler(self.disconnectedCB)


		## login
		self.ph.log("Attempting to log in...", 'connection')

		if self.__con.auth(self.__config['username'], self.__config['password'], self.__config['resource']):
			self.ph.log("Logged in as %s to server %s" % ( self.__config['username'], self.__config['server']), 'connection')
		else:
			return self.ph.log("eek -> %s %s" % (self.__con.lastErr, self.__con.lastErrCode), 'error')

		## roster/presence init
		self.ph.log("Requesting Roster Info" , 'connection')
		self.__con.requestRoster()
		self.__con.sendInitPresence(priority=self.__config['priority'])

		self.JID = self.__config['username'] + '@' + self.__config['server'] + '/' + self.__config['resource']
		return None

	def disconnect(self):
		self.__con.disconnect()


	####################################################################################################
	## input plugin functions

	def process(self):
		try:
			self.__con.process(1)
		except socket.sslerror:
			self.ph._quit()
			
	
	def _send_back(self, msg, message):
		msg = jabber.Message(msg.get_args()['msg'].getFrom(), message)
		msg.setType('chat')
		self.__con.send(msg)


	def _send(self, name, message):
		msg = jabber.Message(name, message)
		msg.setType('chat')
		try:
			self.__con.send(msg)
		except:
			return "error sending jabber message"
		return None
	
	def _is_available(self, id):
		for user in self.__users.keys():
			if user[:len(id)] == id and self.__users[user] == 'available':
				return True
		return False
		
	####################################################################################################
	## command functions
	
	def show_users(self, args):
		all_users = self.__users.keys()
		all_users.sort()

		out = []
		for user in all_users:
			out.append("%s: %s" % (user, self.__users[user]))

		args['msg'].send_back('\n' + join(out, '\n'))

	def reload_users(self, args):
		self.__users = {}
		self.__con.requestRoster()
		args['msg'].send_back('roaster reloaded')

	def __send_presence(self, id, type):
		self.__con.send(jabber.Presence(to=id, type=type))
		
	def send_presence(self, args):
		data = args['cmd'][1:]
		if len(data) != 2:
			args['msg'].send_back('syntax error. see /help')
			return
		id, type = data
		self.__send_presence(id, type)
		args['msg'].send_back("sent %s to %s" % (type, id))
	
	def send_subscribe(self, args):
		data = args['cmd'][1:]
		if len(data) != 1:
			args['msg'].send_back('syntax error. see /help')
			return
		id = data[0]
		self.__send_presence(id, 'subscribe')
		self.__send_presence(id, 'subscribed')
		args['msg'].send_back("subscribed %s" % id)
	
